"""Top-level package for DarkNet OpenSource Neural Networks in Python."""
from ._version import version as __version__  # noqa: F401
from .classifier import Classifier, ImageClassifier
from .detector import ImageDetector

__all__ = ["Classifier", "ImageClassifier", "ImageDetector"]



"""Console script for darknet.py."""

import sys
import click


@click.command()
def py(args=None):
    """Console script for darknet.py."""
    # fmt: off
    click.echo("Replace this message by putting your code into "
               "py.cli.main")
    click.echo("See click documentation at https://click.palletsprojects.com/")
    # fmt: on
    return 0


if __name__ == "__main__":
    sys.exit(py)  # pragma: no cover



from abc import ABC, abstractmethod

import numpy as np

from .network import Network
from .util import image_to_3darray


class ClassifierBase(ABC):
    network: Network
    labels: list

    def __init__(self, labels, config_url, weights_url, **kwargs):
        self.network = Network.open(config_url, weights_url, **kwargs)

        self.labels = range(self.network.output_size()) if labels is None else labels
        if len(self.labels) != self.network.output_size():
            raise TypeError(
                "Number of labels does not match size of network output. "
                f"{len(self.labels)} != {self.network.output_size()}"
            )

    @abstractmethod
    def classify(self, input_obj, top: int = -1):
        pass

    def top_k(self, probabilities, top):
        rv = zip(self.labels, probabilities)
        if top > 0:
            rv = sorted(rv, key=lambda x: x[1], reverse=True)[0:top]
        else:
            rv = list(rv)
        return rv


class Classifier(ClassifierBase):
    def classify(self, input_ndarr: np.ndarray, top: int = -1):
        probabilities = self.network.predict(input)
        return self.top_k(probabilities, top)


class ImageClassifier(ClassifierBase):
    def classify(self, image, top: int = -1):
        image, _ = image_to_3darray(image, self.network.shape)
        probabilities = self.network.predict_image(image)
        return self.top_k(probabilities, top)



import fsspec
import darknet.py.util as darknet_util


def test_fsspec_split_github_url():
    url, kwargs = darknet_util.fsspec_split_github_url("github://org:repo@sha/path/to/file", None)
    assert url == "github://path/to/file"
    assert kwargs["org"] == "org"
    assert kwargs["repo"] == "repo"
    assert kwargs["sha"] == "sha"


def test_fsspec_cache_open(mocker):
    split_spy = mocker.spy(darknet_util, "fsspec_split_github_url")
    fsspec_open_mock = mocker.patch.object(fsspec, "open", autospec=True)

    of = darknet_util.fsspec_cache_open("github://org:repo@sha/path/to/file")
    assert fsspec_open_mock.return_value == of

    fsspec_open_mock.assert_called_once()
    split_spy.assert_called_once()



from .network import Network
from .util import image_to_3darray


class ImageDetector(object):
    network: Network = None
    labels = None

    _last_image_size = None

    def __init__(self, labels, config_url, weights_url, **kwargs):
        self.network = Network.open(config_url, weights_url, **kwargs)
        self.labels = labels

    def detect(self, image, **kwargs):
        self.load_image(image)
        return self.get_detections(**kwargs)

    def load_image(self, image):
        image, self._last_image_size = image_to_3darray(image, self.network.shape)
        self.network.predict_image(image)

    def get_detections(self, **kwargs):
        if "frame_size" not in kwargs:
            kwargs["frame_size"] = self._last_image_size
        detections = self.network.detect(**kwargs)
        return (
            detections
            if self.labels is None
            else [(self.labels[label_idx], prob, bbox) for label_idx, prob, bbox in detections]
        )



# network.pyx
from copy import deepcopy

import numpy as np
cimport numpy as np
cimport libdarknet as dn

from libc.stdlib cimport free
from .util import fsspec_cache_open

np.import_array()

cdef class Metadata:
    classes = []  # typing: List[AnyStr]

    def __cinit__(self, config_file):
        cdnet_mdata = dn.get_metadata(config_file.encode())
        for i in range(cdnet_mdata.classes):
            self.classes.append(cdnet_mdata.names[i].decode("utf-8"))

    def get_classes(self):
        return deepcopy(self.classes)


cdef class Network:
    cdef dn.network* _c_network

    @staticmethod
    def open(config_url, weights_url):
        with fsspec_cache_open(config_url, mode="rt") as config:
            with fsspec_cache_open(weights_url, mode="rb") as weights:
                return Network(config.name, weights.name)


    def __cinit__(self, config_file, weights_file):
        clear = 1
        self._c_network = dn.load_network(config_file.encode(), weights_file.encode(), clear)
        if self._c_network is NULL:
            raise RuntimeError("Failed to create the DarkNet Network...")

    def __dealloc__(self):
        if self._c_network is not NULL:
            dn.free_network(self._c_network[0])
            free(self._c_network)

    @property
    def shape(self):
        return dn.network_width(self._c_network), dn.network_height(self._c_network)

    def input_size(self):
        return dn.network_input_size(self._c_network)

    def output_size(self):
        return dn.network_output_size(self._c_network)

    def predict(self, np.ndarray[dtype=np.float32_t, ndim=1, mode="c"] input) -> np.ndarray:
        cdef int input_size
        input_size = self.input_size()
        if input.size != input_size:
            raise TypeError("The input array size does not match the network input size. "
                            f"({input.size} != {input_size})")

        cdef float* output
        output = dn.network_predict(self._c_network[0], <float *>input.data)

        cdef np.npy_intp output_shape[1]
        output_shape[0] = self.output_size()
        return np.PyArray_SimpleNewFromData(1, output_shape, np.NPY_FLOAT32, output)

    def predict_image(self, np.ndarray[dtype=np.float32_t, ndim=3, mode="c"] img) -> np.ndarray:
        cdef dn.image imr
        imr.c = img.shape[0]
        imr.h = img.shape[1]
        imr.w = img.shape[2]
        imr.data = <float *> img.data

        cdef float* output
        output = dn.network_predict_image(self._c_network, imr)

        cdef np.npy_intp output_shape[1]
        output_shape[0] = self.output_size()
        return np.PyArray_SimpleNewFromData(1, output_shape, np.NPY_FLOAT32, output)

    def detect(self, frame_size=None,
               float threshold=.5, float hierarchical_threshold=.5,
               int relative=0, int letterbox=1,
               str nms_type="sort", float nms_threshold=.45,
               ):
        frame_size = self.shape if frame_size is None else frame_size
        cdef int num_dets = 0
        cdef dn.detection* detections

        detections = dn.get_network_boxes(self._c_network,
                                          frame_size[0], frame_size[1],
                                          threshold, hierarchical_threshold,
                                          <int*>0,
                                          relative,
                                          &num_dets,
                                          letterbox)

        if nms_threshold > 0 and num_dets:
            if nms_type == "obj":
                dn.do_nms_obj(detections, num_dets, detections[0].classes, nms_threshold)
            elif nms_type == "sort":
                dn.do_nms_sort(detections, num_dets, detections[0].classes, nms_threshold)
            else:
                raise ValueError(f"non-maximum-suppression type {nms_type} is not one of {['obj', 'sort']}")

        rv = [
            (j, detections[i].prob[j],
             (detections[i].bbox.x, detections[i].bbox.y, detections[i].bbox.w, detections[i].bbox.h))
            for i in range(num_dets)
            for j in range(detections[i].classes)
            if detections[i].prob[j] > 0
        ]

        dn.free_detections(detections, num_dets)
        return sorted(rv, key=lambda x: x[1], reverse=True)



import os
import fsspec

import numpy as np
from PIL import Image
from PIL import ImageDraw


def fsspec_cache_open(
    urlpath: str,
    mode="rb",
    compression=None,
    encoding="utf8",
    errors=None,
    protocol=None,
    newline=None,
    **kwargs,
) -> fsspec.core.OpenFile:
    chain = urlpath.split("::")

    if chain[0].startswith("github"):
        chain[0], kwargs = fsspec_split_github_url(chain[0], kwargs)

    # Because darknet is written in C, we need real file names for it to open
    if chain[0] not in {"filecache", "simplecache"}:
        first_scheme = chain[0].split("://")[0]
        urlpath = f"filecache::{urlpath}"
        filecache = dict(cache_storage=f"{os.environ['HOME']}/.cache/darknet.py")
        kwargs = {"filecache": filecache, first_scheme: kwargs}

    return fsspec.open(urlpath, mode, compression, encoding, errors, protocol, newline, **kwargs)


def fsspec_split_github_url(github_url: str, kwargs: dict) -> (str, dict):
    # TODO: Remove this once fsspec > 0.7.5
    from urllib.parse import urlparse

    rv = github_url
    github_url = urlparse(github_url)
    keys = {"org", "repo", "sha"}
    # If that metadata is not passed as kwargs, we need to extract it
    # netloc = "{org}:{repo}@{sha}"
    kwargs = kwargs or dict()
    if (keys & kwargs.keys()) != keys:
        org, repo, sha = github_url.username, github_url.password, github_url.hostname
        if org is None or repo is None or sha is None:
            raise ValueError(
                f"The github url {github_url} does not match `github://<org>:<repo>@<sha>/path`"
            )
        kwargs.update(dict(org=org, repo=repo, sha=sha))
        rv = github_url.geturl().replace(f"{github_url.netloc}/", "")

    return rv, kwargs


def image_to_3darray(image, target_shape):
    # We assume the original size matches the target_shape (height, width)
    orig_size = target_shape

    if isinstance(image, str):
        with fsspec.open(image, mode="rb") as f:
            image = Image.open(f)

    if isinstance(image, Image.Image):
        orig_size = image.size
        image = image_scale_and_pad(image, target_shape)
        image = np.asarray(image)

    if isinstance(image, np.ndarray):
        if image.shape[0:2] != target_shape:
            image = Image.fromarray(image)
            orig_size = image.size
            image = image_scale_and_pad(image, target_shape)
            image = np.asarray(image)
        image = image.transpose((2, 0, 1)).astype(dtype=np.float32, order="C") / 255
    return image, orig_size


def image_scale_and_pad(image: Image.Image, target_shape) -> Image.Image:
    image = image.convert("RGB")
    if (image.width, image.height) != target_shape:
        from PIL import ImageOps

        image = image.copy()
        image.thumbnail(target_shape)
        image = ImageOps.pad(image, target_shape)
    return image


def image_draw_detections(img: Image.Image, detections) -> Image.Image:
    img = img.copy()
    draw = ImageDraw.Draw(img)
    colors = ["purple", "blue", "green", "pink", "brown"]

    def xywh_to_bounds(x, y, w, h):
        return x - w / 2, y - h / 2, x + w / 2, y + h / 2

    for i, (cat, prob, xywh) in enumerate(detections):
        text = f"{cat}@{prob:.2%}"
        bounds = xywh_to_bounds(*xywh)
        t_w, t_h = draw.textsize(text)
        draw.rectangle(xywh_to_bounds(*xywh), outline=colors[i % 5], width=4)
        draw.rectangle(
            (bounds[0], bounds[1] - t_h, bounds[0] + t_w + 4, bounds[1]), fill=colors[i % 5]
        )
        draw.text((bounds[0] + 2, bounds[1] - t_h), text, fill="white")
    return img



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_namespace_packages, Extension
import os

with open("README.rst") as readme_file:
    readme = readme_file.read()

# The requirements section should be kept in sync with the environment.yml file
requirements = [
    # fmt: off
    "click>=7.0",
    "click-plugins",
    "entrypoints",
    "fsspec",
    "numpy",
    "pillow",
    # fmt: on
]

setup_requirements = [
    # fmt: off
    "cython",
    "numpy",
    "setuptools_scm",
    "setuptools_scm_git_archive",
    "wheel",
    # fmt: on
]

test_requirements = [
    # fmt: off
    "pytest>=3",
    "pytest-cov",
    "pytest-mock",
    # fmt: on
]

doc_requirements = [
    # fmt: off
    "sphinx",
    "sphinx-autoapi",
    "sphinx-click",
    "watchdog",
    # fmt: on
]

zoo_requirements = [
    # fmt: off
    "intake",
    # fmt: on
]

conda_requires = {
    # fmt: off
    "build": [
    ],
    "host": [
        "darknet-cpu >=0.0.20200614"
    ],
    "run": [
        "darknet >=0.0.20200614"
    ],
    "test": [
        "darknet-cpu >=0.0.20200614"
        "intake"
    ],
    # fmt: on
}

conda_rosetta_stone = {
    # fmt: off
    "pypa-requirement": "conda-dependency"
    # fmt: on
}

ext_modules = []
try:
    from Cython.Build import cythonize
    import numpy as np

    ext_modules = cythonize(
        [
            Extension(
                "darknet.py.network",
                ["src/darknet/py/network.pyx"],
                include_dirs=[np.get_include()],
                libraries=["darknet"],
            )
        ]
    )
except ModuleNotFoundError:
    pass

setup_kwargs = dict(
    author="Patrick Sodré",
    author_email="psodre@gmail.com",
    use_scm_version={"write_to": "src/darknet/py/_version.py"},
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="DarkNet OpenSource Neural Networks in Python",
    # fmt: off
    entry_points={
        "darknet.cli": [
            "py=darknet.py.cli:py",
        ],
        "intake.drivers": [
            "darknet = darknet.py.intake:DarknetSource"
        ],
        "intake.catalogs": [
            "darknet = darknet.zoo:cat"
        ]
    },
    # fmt: on
    ext_modules=ext_modules,
    install_requires=requirements,
    license="MIT",
    long_description=readme,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="py darknet",
    name="darknet-py",
    package_dir={"": "src"},
    packages=find_namespace_packages(where="./src"),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    extras_require={
        # fmt: off
        "test": test_requirements,
        "doc": doc_requirements,
        "zoo": zoo_requirements,
        # fmt: on
    },
    url="https://github.com/zeroae/darknet.py",
    zip_safe=False,
)

if "CONDA_BUILD_STATE" in os.environ:
    try:
        from setuptools_scm import get_version

        setup_kwargs["version"] = get_version(**setup_kwargs["use_scm_version"])
        del setup_kwargs["use_scm_version"]
    except ModuleNotFoundError:
        print(
            "Error: darknet.py requires that setuptools_scm be installed with conda-build!"  # noqa: E501
        )
        raise
    setup_kwargs["conda_rosetta_stone"] = conda_rosetta_stone
    setup_kwargs["conda_requires"] = conda_requires

setup(**setup_kwargs)



from typing import Iterable

import fsspec
from intake.source.base import DataSource, Schema

from darknet.py import ImageClassifier


class DarknetSource(DataSource):
    name = "darknet"
    version = "0.0.1"
    container = "python"
    partition_access = False

    def __init__(
        self, names, net_config, net_weights, names_slice=None, storage_options=None, metadata=None
    ):
        self._names = names
        names_slice = names_slice or (-1,)
        names_slice = names_slice if isinstance(names_slice, Iterable) else (names_slice,)
        self._names_slice = slice(*names_slice)
        self._net_config = net_config
        self._net_weights = net_weights
        self._storage_options = storage_options or {}
        super().__init__(storage_options, metadata)

    def _get_schema(self):
        return Schema(
            datashape=None, dtype=None, shape=(None,), npartitions=0, extra_metadata=self.metadata
        )

    def _get_partition(self, i):
        with fsspec.open(self._names, mode="rt", encoding="utf-8", **self._storage_options) as f:
            names = [line.rstrip() for line in f.readlines()][self._names_slice]
        with fsspec.open(
            self._net_config, mode="rt", encoding="utf-8", **self._storage_options
        ) as net_config:
            with fsspec.open(self._net_weights, mode="rb", **self._storage_options) as net_weights:
                return ImageClassifier(names, net_config.name, net_weights.name)



import intake
import os

here = os.path.abspath(os.path.dirname(__file__))
zoo = intake.open_catalog(os.path.join(here, "darknet.yaml"))
imagenet = intake.open_catalog(os.path.join(here, "imagenet.yaml"))

del here



# __init__.pxd

cdef extern from "darknet.h":
    """
    /*
     * darknet.h forgot to extern some useful network functions
     */
     static int network_input_size(network* net) {
        return net->layers[0].inputs;
     }
     static int network_output_size(network* net) {
        int i;
        for(i = net->n-1; i > 0; --i) if(net->layers[i].type != COST) break;
        return net->layers[i].outputs;
     }
    """

    # The input vector
    # An RGB or Gray-scale image.
    ctypedef struct image:
        int w;
        int h;
        int c;
        float* data;

    void free_image(image self)

    # The output vector
    ctypedef struct metadata:
        int classes
        char** names

    metadata get_metadata(char* filename)

    # Detection Output
    ctypedef struct box:
        float x;
        float y;
        float w;
        float h;

    ctypedef struct detection:
        box bbox;
        int classes;
        float* prob;
        float* mask;
        float objectness;
        int sort_class;
        float *uc;  # Gaussian_YOLOv3 - tx, ty, tw, th uncertainty
        int points; # bit-0 - center, bit-1 top-left-corner, bit-2 bottom-right-corner

    void free_detections(detection* detections, int len)

    void do_nms_sort(detection* detections, int len, int num_classes, float thresh)
    void do_nms_obj(detection* detections, int len, int num_classes, float thresh)

    # The model
    ctypedef struct network:
        pass

    network* load_network(char* cfg_filename, char* weights_filename, int clear)
    void free_network(network self)

    int network_width(network *self);
    int network_height(network *self);
    int network_input_size(network* self);
    int network_output_size(network* self);
    float* network_predict(network, float* input)
    float* network_predict_image(network*, image)
    detection* get_network_boxes(network* self, int width, int height, float thresh, float hier_thresh, int* map, int relative, int* out_len, int letter)


